//////////////////////////////////////////////////////////////////////////
//
//			VIEWING VOLUME VISUALIZER
//	A didactical tool to understand viewing and projection in OpenGL,
//			using C, C++ and the GLUT and GLUI libraries 
//
//////////////////////////////////////////////////////////////////////////
//  Thomas W. Rauber http://www.inf.ufes.br/~thomas
//	with the collaboration of
//    Thiago Oliveira dos Santos
//////////////////////////////////////////////////////////////////////////


#ifndef FEEDBACK
extern void create_glui_interface(); 
#endif

#ifdef FEEDBACK
extern /* "C" */ void initEpsdriver(void);
extern /* "C" */ void displayFeedback(void);
extern /* "C" */ void keyboard( unsigned char key, int x, int y );
#endif

#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "glutobj.h"
#include "mvproj.h"



static GLdouble mv[16], mp[16]; /* MODELVIEW e PROJECTION MATRIX */
static GLint vp[4];     /* VIEWPORT */

//#define VERTICAL // Draw the second observer below or right of main observer
#ifdef VERTICAL
#define w_init 300
#define h_init 600
#endif
#ifndef VERTICAL
#define w_init 600
#define h_init 300
#endif
static GLint w, h;	/* Width and height of window for reshaping */

#ifndef FEEDBACK
int main_window;
#endif


/* The Principal View (what the main camera sees): mapped to upper
	half of the window */
float left, right, bottom, top, near_, far_;

float eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz;

/* if 'gluPerspective is used in perspective projection */
float fovy, aspect;

int orthographic;  /* using orthographic proj. */
int frustum;  /* if perspective proj. use glFrustum */


/* Meta-View: A second camera observes the first observer,
	together with his viewing volume */
float left2, right2, bottom2, top2, near_2, far_2;

float eyex2, eyey2, eyez2, centerx2, centery2, centerz2,
			upx2, upy2, upz2;

/* if 'gluPerspective is used in perspective projection */
float fovy2, aspect2;

int orthographic2; /* using orthographic proj. */
int frustum2; /* if perspective proj. use glFrustum */



/* Reset to initial conditions of the system */
void Reset(int i = 0)
{
  int j, k;

  //camera 1
  eyex = 0.0; eyey = 0.0; eyez = 5.0;
  centerx = 0.0; centery = 0.0; centerz = 0.0;
  upx = 0.0; upy = 1.0; upz = 0.0;
  left = -1.5; right = 1.5; bottom = -1.5; top = 1.5;
  near_ = 3.0; far_ = 6.5;
  fovy = 20.0; aspect = 4.0/3.0;
  orthographic = 0; 
  frustum = 1;  

  //camera 2
  eyex2 = 20.0; eyey2 = 10.0; eyez2 = 10.0; 
  centerx2 = 0.0; centery2 = 0.0; centerz2 = 0.0; 
  upx2 = 0.0; upy2 = 1.0; upz2 = 0.0;
  left2 = -2.0; right2 = 2.0; bottom2 = -2.0; top2 = 2.0; 
  //near_2 = 5.0; far_2 = 20.0;
  //eyex2 = 15.0; eyey2 = 0.0; eyez2 = 0.0; 
  near_2 = 8.0; far_2 = 40.0;
  fovy2 = 20.0; aspect2 = 4.0/3.0;
  orthographic2 = 0;
  frustum2 = 1; 

  animation_x_z = 0; animation_x_y = 0;
  //reset the array of the object rotation control
  for( j=0; j < 4; j++ )
    for( k=0; k < 4; k++ )
      if( j == k )
        view_rotate[j*4+k] = 1.0;
      else
        view_rotate[j*4+k] = 0.0;

  delay = 10;

  // Object parameters
  dSize = 1.0;
  radius=1.0; slices=30; stacks=10;
  base=1.0; height=1.0;
  dInnerRadius=0.5; dOuterRadius=1.0;
  nSides=30; nRings=20;
  num_levels=1; offset[0]=offset[1]=offset[2]=0.0; scale=1.0;

  glutReshapeWindow( w_init, h_init );
}


/************************************************** draw_axes() **********/
/* Disables lighting, then draws RGB axes                                */

static void draw_axes( float scale )
{
  glPushMatrix();
  glScalef( scale, scale, scale );

  //glPushAttrib( GL_LINE_BIT | GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT );
  //glPushAttrib( GL_LINE_BIT | GL_COLOR_BUFFER_BIT );
  glPushAttrib( GL_ALL_ATTRIB_BITS );
  glDisable( GL_LIGHTING );
  glPointSize( 1.0 );
  glBegin( GL_LINES );
  glColor3f( 1.0, 0.0, 0.0 );
  glVertex3f( 0.8f, 0.05f, 0.0 ); glVertex3f( 1.0f, 0.25f, 0.0 ); /* Letter X */
  glVertex3f( 0.8f, 0.25f, 0.0 ); glVertex3f( 1.0f, 0.05f, 0.0 );
  glVertex3f( 0.0, 0.0, 0.0 ); glVertex3f( 1.0f, 0.0, 0.0 ); /* X axis      */

  glColor3f( 0.0, 1.0, 0.0 );
  glVertex3f( 0.0, 0.8f, 0.25f ); glVertex3f( 0.0, 0.9f, 0.15f ); /* Letter Y */
  glVertex3f( 0.0, 1.0f, 0.25f ); glVertex3f( 0.0, 0.9f, 0.15f );
  glVertex3f( 0.0, 0.9f, 0.15f ); glVertex3f( 0.0, 0.9f, 0.05f );
  glVertex3f( 0.0, 0.0, 0.0 );  glVertex3f( 0.0, 1.0, 0.0 ); /* Y axis      */

  glColor3f( 0.0, 0.0, 1.0 );
  glVertex3f( 0.25f, 0.0, 0.8f ); glVertex3f( 0.25f, 0.0, 1.0f ); /* Letter Z */
  glVertex3f( 0.25f, 0.0, 1.0f ); glVertex3f( 0.05f, 0.0, 0.8f );
  glVertex3f( 0.05f, 0.0, 0.8f ); glVertex3f( 0.05f, 0.0, 1.0f );
  glVertex3f( 0.0, 0.0, 0.0 );  glVertex3f( 0.0, 0.0, 1.0 ); /* Z axis    */
  glEnd();

  glPopMatrix();
  glEnable( GL_LIGHTING );
  glPopAttrib();
}

/*
 * Given the viewport coordinates, calculate the 8 points that define
 * the viewing volume by unprojecting the window coordinates back
 * to world coordinates using the modelview and projection matrix
 * and the viewport of the view that was defined by the viewing volume
 * (principal view)
*/
/* The vertex indices for the planes */
static GLubyte near_Indices[4]	=	{ 0, 1, 2, 3 };
static GLubyte far_Indices[4]	=	{ 4, 5, 6, 7 };
static GLubyte leftIndices[4]	=	{ 4, 0, 3, 7 };
static GLubyte rightIndices[4]	=	{ 5, 1, 2, 6 };
static GLubyte buttomIndices[4]	=	{ 0, 1, 5, 4 };
static GLubyte topIndices[4]	=	{ 3, 2, 6, 7 };
static GLubyte allIndices[8]	=	{ 0, 1, 2, 3, 4, 5, 6, 7 };


#define D 3
static GLdouble xmin, xmax, ymin, ymax, zmin = 0.0, zmax = 1.0;
static GLdouble winx, winy, winz, objx, objy, objz, vv[8*D];
static GLfloat alpha = 0.3;

static void draw_viewing_volume()
{
  xmin = (GLdouble)(vp[0]); xmax = (GLdouble)(vp[0]+vp[2]); 
  ymin = (GLdouble)(vp[1]); ymax = (GLdouble)(vp[1]+vp[3]); 

  /* Desabilita a luz enquanto desenha o volume */
  glDisable( GL_LIGHTING );  

  /* left, button, near_ */
  winx = xmin; winy = ymin; winz = zmin;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[0*D+0] = objx; vv[0*D+1] = objy; vv[0*D+2] = objz;

  /* right, button, near_ */
  winx = xmax; winy = ymin; winz = zmin;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[1*D+0] = objx; vv[1*D+1] = objy; vv[1*D+2] = objz;

  /* right, top, near_ */
  winx = xmax; winy = ymax; winz = zmin;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[2*D+0] = objx; vv[2*D+1] = objy; vv[2*D+2] = objz;

  /* left, top, near_ */
  winx = xmin; winy = ymax; winz = zmin;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[3*D+0] = objx; vv[3*D+1] = objy; vv[3*D+2] = objz;

  /* left, button, far_ */
  winx = xmin; winy = ymin; winz = zmax;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[4*D+0] = objx; vv[4*D+1] = objy; vv[4*D+2] = objz;

  /* right, button, far_ */
  winx = xmax; winy = ymin; winz = zmax;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[5*D+0] = objx; vv[5*D+1] = objy; vv[5*D+2] = objz;

  /* right, top, far_ */
  winx = xmax; winy = ymax; winz = zmax;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[6*D+0] = objx; vv[6*D+1] = objy; vv[6*D+2] = objz;

  /* left, top, far_ */
  winx = xmin; winy = ymax; winz = zmax;
  gluUnProject( winx, winy, winz, mv, mp, vp, &objx, &objy, &objz );
  vv[7*D+0] = objx; vv[7*D+1] = objy; vv[7*D+2] = objz;

  /* Have eight vertices of the viewing volume, compose sides
  by transparent polygons */

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(D, GL_DOUBLE, 0, vv);

  /* All vertices of viewing volumes as points */
  glPointSize( 2.0 );
  glColor3f( 1.0, 0.0, 0.0 );
  glDrawElements( GL_POINTS, 8, GL_UNSIGNED_BYTE, allIndices );

  if ( vvwireframe )
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  glColor4f( 1.0, 0.0, 0.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, near_Indices );
  glColor4f( 0.0, 1.0, 0.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, far_Indices );
  glColor4f( 0.0, 0.0, 1.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, leftIndices );
  glColor4f( 1.0, 1.0, 0.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, rightIndices );
  glColor4f( 1.0, 0.0, 1.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, buttomIndices );
  glColor4f( 0.0, 1.0, 1.0, alpha );
  glDrawElements( GL_QUADS, 4, GL_UNSIGNED_BYTE, topIndices );

  glDisable(GL_BLEND);
  glDisableClientState(GL_VERTEX_ARRAY);

  glEnable( GL_LIGHTING );  
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

static void draw_eye_cam()
{
  GLdouble camlb[3], camrb[3], camrt[3], camlt[3],camrtu[3], camltu[3], scal;

  scal = (GLdouble)(0.10 * near_);
  camlb[0] = eyex + scal * (vv[0*D+0] - eyex);
  camlb[1] = eyey + scal * (vv[0*D+1] - eyey);
  camlb[2] = eyez + scal * (vv[0*D+2] - eyez);
  camrb[0] = eyex + scal * (vv[1*D+0] - eyex);
  camrb[1] = eyey + scal * (vv[1*D+1] - eyey);
  camrb[2] = eyez + scal * (vv[1*D+2] - eyez);
  camrt[0] = eyex + scal * (vv[2*D+0] - eyex);
  camrt[1] = eyey + scal * (vv[2*D+1] - eyey);
  camrt[2] = eyez + scal * (vv[2*D+2] - eyez);
  camlt[0] = eyex + scal * (vv[3*D+0] - eyex);
  camlt[1] = eyey + scal * (vv[3*D+1] - eyey);
  camlt[2] = eyez + scal * (vv[3*D+2] - eyez);
  camrtu[0] = eyex + scal * 1.5 * (vv[2*D+0] - eyex);
  camrtu[1] = eyey + scal * 1.5 * (vv[2*D+1] - eyey);
  camrtu[2] = eyez + scal * 1.5 * (vv[2*D+2] - eyez);
  camltu[0] = eyex + scal * 1.5 * (vv[3*D+0] - eyex);
  camltu[1] = eyey + scal * 1.5 * (vv[3*D+1] - eyey);
  camltu[2] = eyez + scal * 1.5 * (vv[3*D+2] - eyez);

  //printf("Eye  =(%12.9lf,%12.9lf,%12.9lf)\n", eyex, eyey, eyez );/**/
  glPushMatrix();
  glColor3f( 0.0, 0.0, 0.0 );
  glBegin(GL_LINE_LOOP);
    glVertex3dv( camlb );
    glVertex3dv( camrb );
    glVertex3dv( camrt );
    glVertex3dv( camlt );
  glEnd();    
  glBegin(GL_LINES);
    glVertex3d( eyex, eyey, eyez );
    glVertex3dv( camlb );
    glVertex3d( eyex, eyey, eyez );
    glVertex3dv( camrb );
    glVertex3d( eyex, eyey, eyez );
    glVertex3dv( camrt );
    glVertex3d( eyex, eyey, eyez );
  glVertex3dv( camlt );
  glEnd();

  // Draw Up direction of camera
  glBegin(GL_LINES);
    glVertex3dv( camltu );
    glVertex3dv( camlt );
    glVertex3dv( camrtu );
    glVertex3dv( camrt );
    glVertex3dv( camrtu );
    glVertex3dv( camltu );
  glEnd();
  glPopMatrix();

  /* Draw eye position as red thick point */
  glPushAttrib( GL_LINE_BIT | GL_COLOR_BUFFER_BIT );
  glDisable( GL_LIGHTING );
  glPointSize( 3.0 );
  glColor3f( 1.0, 0.0, 0.0 );
  glBegin(GL_POINTS); glVertex3d( eyex, eyey, eyez ); glEnd();
  glEnable( GL_LIGHTING );
  glPopAttrib();
} 

// Draw your own object
static void draw_my_object()
{
  glPushMatrix();
  glMultMatrixf( view_rotate );

  draw_axes( 1.0 );
  glTranslatef( 1.2, 0.0, 0.0 );
  draw_axes( 1.0 );
  glRotatef( 30.0, 0.0, 0.0, 1.0 );
  draw_axes( 1.0 );

  glPopMatrix();
}

static void draw_object()
{
  /* draw_my_object(); return; /**/

  glPushMatrix();

  // Rotate
  glMultMatrixf( view_rotate );

  draw_axes( 1.0 );

  // Select type of object
  if ( wireframe )
    switch( object_type ) {
      case 0: my_freeglutWireTeapot( (GLdouble)dSize ); break;
      case 1: my_freeglutWireCube( (GLdouble)dSize ); break;
      case 2: my_freeglutWireSphere( (GLdouble)radius, (GLint)slices, (GLint)stacks ); break;
      case 3: my_freeglutWireCone( (GLdouble)base, (GLdouble)height, (GLint)slices, (GLint)stacks ); break;
      case 4: my_freeglutWireCylinder( (GLdouble)radius, (GLdouble)height, (GLint)slices, (GLint)stacks ); break;
      case 5: my_freeglutWireTorus( (GLdouble)dInnerRadius, (GLdouble)dOuterRadius, (GLint)nSides, (GLint)nRings ); break;
      case 6: my_freeglutWireDodecahedron(); break;
      case 7: my_freeglutWireOctahedron(); break;
      case 8: my_freeglutWireTetrahedron(); break;
      case 9: my_freeglutWireIcosahedron(); break;
      case 10: my_freeglutWireRhombicDodecahedron(); break;
      case 11: my_freeglutWireSierpinskiSponge ( num_levels, offset, (GLdouble)scale ); break;
      default: fprintf(stderr,"Wrong object type. Exit...\n"); exit(1);
    }
  else
    switch( object_type ) {
      case 0: my_freeglutSolidTeapot( (GLdouble)dSize ); break;
      case 1: my_freeglutSolidCube( (GLdouble)dSize ); break;
      case 2: my_freeglutSolidSphere( (GLdouble)radius, (GLint)slices, (GLint)stacks ); break;
      case 3: my_freeglutSolidCone( (GLdouble)base, (GLdouble)height, (GLint)slices, (GLint)stacks ); break;
      case 4: my_freeglutSolidCylinder( (GLdouble)radius, (GLdouble)height, (GLint)slices, (GLint)stacks ); break;
      case 5: my_freeglutSolidTorus( (GLdouble)dInnerRadius, (GLdouble)dOuterRadius, (GLint)nSides, (GLint)nRings ); break;
      case 6: my_freeglutSolidDodecahedron(); break;
      case 7: my_freeglutSolidOctahedron(); break;
      case 8: my_freeglutSolidTetrahedron(); break;
      case 9: my_freeglutSolidIcosahedron(); break;
      case 10: my_freeglutSolidRhombicDodecahedron(); break;
      case 11: my_freeglutSolidSierpinskiSponge ( num_levels, offset, (GLdouble)scale ); break;
      default: fprintf(stderr,"Wrong object type. Exit...\n"); exit(1);
    }

  glPopMatrix();
}


void show_matrix(double *M, char *comment)
{
  printf("\n\t\%s\n", comment);
  printf("\t%f, %f, %f, %f\n", M[ 0],M[ 4],M[ 8],M[12]);
  printf("\t%f, %f, %f, %f\n", M[ 1],M[ 5],M[ 9],M[13]);
  printf("\t%f, %f, %f, %f\n", M[ 2],M[ 6],M[10],M[14]);
  printf("\t%f, %f, %f, %f\n", M[ 3],M[ 7],M[11],M[15]);
  printf("\n\n");
}


void show_current_modelview_matrix(char *comment)
{
  GLdouble mv[16];

  glGetDoublev( GL_MODELVIEW_MATRIX, mv );
  printf("\tCURRENT MODELVIEW MATRIX\n\t\%s\n", comment);
  printf("\t%f, %f, %f, %f\n", mv[ 0],mv[ 4],mv[ 8],mv[12]);
  printf("\t%f, %f, %f, %f\n", mv[ 1],mv[ 5],mv[ 9],mv[13]);
  printf("\t%f, %f, %f, %f\n", mv[ 2],mv[ 6],mv[10],mv[14]);
  printf("\t%f, %f, %f, %f\n", mv[ 3],mv[ 7],mv[11],mv[15]);
  printf("\n\n");
}


extern GLboolean invert_matrix(const GLdouble * m, GLdouble * out);
static GLdouble mv1inv[16]; // Inverse modelview matrix of view 1
/* Draw the eye coordinate axes */
static void draw_eye_coordsys_axes()
{
  //show_matrix(mv,"Before inversion");
  GLboolean ok = invert_matrix( mv, mv1inv );
  //show_matrix(mv1inv,"After inversion mv1inv");
 
  glPushMatrix();

  // Transform from object coordinates to eye coordinate system of view 1
  glMultMatrixd( mv1inv );

  draw_axes( 1.0 );

  glPopMatrix();
}


void update_mv1()
{
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();
  gluLookAt( eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz );
  glGetDoublev( GL_MODELVIEW_MATRIX, mv );
  glPopMatrix();
}


/* Draw scene from view 1 into viewport 1 */
static void draw_view1()
{
  /* Specify projection: Perspective or Orthographic */
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  if( orthographic )
    glOrtho( left, right, bottom, top, near_, far_ );
  else    /* perspective projection */
    if( frustum )
     glFrustum( left, right, bottom, top, near_, far_ );
    else
     gluPerspective( fovy, aspect, near_, far_ );

  /* Specify view */
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  //show_current_modelview_matrix("After glLoadIdentity()");
  gluLookAt( eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz );
  //show_current_modelview_matrix("After gluLookAt()");

  /* Specify viewport */
#ifdef VERTICAL
  glViewport( 0, 0, w, h/2 );
#endif
#ifndef VERTICAL
  glViewport( 0, 0, w/2, h );
#endif

  /* Save viewing and projection of first view */
  glGetDoublev( GL_MODELVIEW_MATRIX, mv );
  glGetDoublev( GL_PROJECTION_MATRIX, mp );
  glGetIntegerv( GL_VIEWPORT, vp );

  draw_object();
}

/* Draw scene from view 2 into viewport 2 */
static void draw_view2()
{
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  /* Specify projection */
  if( orthographic2 )
    glOrtho( left2, right2, bottom2, top2, near_2, far_2 );
  else    /* perspective projection */
    if( frustum2 )
      glFrustum( left2, right2, bottom2, top2, near_2, far_2 );
    else
      gluPerspective( fovy2, aspect2, near_2, far_2 );

  /* Specify view */
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  gluLookAt( eyex2, eyey2, eyez2, centerx2, centery2, centerz2,
		upx2, upy2, upz2);

  /* Specify viewport */
#ifdef VERTICAL
  glViewport( 0, h/2, w, h/2 );
#endif
#ifndef VERTICAL
  glViewport( w/2, 0, w/2, h );
#endif

  draw_object();
  draw_eye_coordsys_axes();
  draw_viewing_volume();
  draw_eye_cam(); 
}

static char codefilename[] = "_specify_viewing_and_projection.c";
void gen_c_code_viewing_and_proj( int i )
{
  FILE *p;

  p = fopen( codefilename, "w" );
  if (p==NULL){
    printf("\n\n\tUnable to open code file '%s'!", codefilename );
    return;
  }

  fprintf( p, "void specify_viewing_and_projection()\n" );
  fprintf( p, "{\n" );
  fprintf( p, "  /* Specify modelview matrix */\n" );
  fprintf( p, "  glMatrixMode( GL_MODELVIEW );\n" );
  fprintf( p, "  glLoadIdentity();\n" );
  fprintf( p, "  gluLookAt( %f, %f, %f,\n\t%f, %f, %f,\n\t%f, %f, %f );\n",
	eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz );
  fprintf( p, "\n  /* Specify projection matrix */\n" );
  fprintf( p, "  glMatrixMode( GL_PROJECTION );\n" );
  fprintf( p, "  glLoadIdentity();\n" );
  if( orthographic )
    fprintf( p, "  glOrtho( %f, %f, %f, %f, %f, %f);\n",
			left, right, bottom, top, near_, far_);
  else
    if( frustum )
      fprintf( p, "  glFrustum( %f, %f, %f, %f, %f, %f);\n",
			left, right, bottom, top, near_, far_);
    else
      fprintf( p, "  gluPerspective( %f, %f, %f, %f);\n",
			fovy, aspect, near_, far_);
  // From now on manipulate modelview matrix
  fprintf( p, "\n  glMatrixMode( GL_MODELVIEW );\n" );
  fprintf( p, "}\n" );
  fclose(p);
  printf("\nSuccessfully generated C-code in file '%s'\n", codefilename );
}

void show_modelview_and_projection_matrices(int dummy)
{
  printf("\n--- MODELVIEW, PROJECTION AND VIEWPORT OF PRINCIPAL VIEW ---\n\n");
  printf("\tMODELVIEW MATRIX\n\n");
  printf("\t%f, %f, %f, %f\n", mv[ 0],mv[ 4],mv[ 8],mv[12]);
  printf("\t%f, %f, %f, %f\n", mv[ 1],mv[ 5],mv[ 9],mv[13]);
  printf("\t%f, %f, %f, %f\n", mv[ 2],mv[ 6],mv[10],mv[14]);
  printf("\t%f, %f, %f, %f\n", mv[ 3],mv[ 7],mv[11],mv[15]);

  printf("\n\n");

  printf("\tPROJECTION MATRIX\n\n");
  printf("\t%f, %f, %f, %f\n", mp[ 0],mp[ 4],mp[ 8],mp[12]);
  printf("\t%f, %f, %f, %f\n", mp[ 1],mp[ 5],mp[ 9],mp[13]);
  printf("\t%f, %f, %f, %f\n", mp[ 2],mp[ 6],mp[10],mp[14]);
  printf("\t%f, %f, %f, %f\n", mp[ 3],mp[ 7],mp[11],mp[15]);

  printf("\n\n");

  printf("\tVIEWPORT \n\n");
  printf("\txmin = %d\n", vp[0]);
  printf("\tymin = %d\n", vp[1]);
  printf("\txmax = %d\n", vp[0]+vp[2]);
  printf("\tymax = %d\n", vp[1]+vp[3]);
}


static void printstring(void *font, char *string)
{
   int len, i;

   len = (int) strlen(string);
   for (i = 0; i < len; i++)
      glutBitmapCharacter(font, string[i]);
}

/* Draw a horizontal line between the two viewports of the main window
   and some explanatory text */
static void draw_info()
{
  /* save the current view and projection */
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho( -1.0, 1.0, -1.0, 1.0, -1.0, 1.0 );
  glViewport( 0, 0, w, h );
  glPushAttrib( GL_LINE_BIT | GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT );
  //glPushAttrib( GL_ALL_ATTRIB_BITS );
  glDisable(GL_LIGHTING);
  glPointSize( 1.0 );
  glColor3f( 0.0, 0.0, 0.0 );
#ifdef FEEDBACK	// Draw also box around the main window
  glBegin(GL_LINE_LOOP); glVertex3f(-1.0,-1.0,0.0); glVertex3f(1.0,-1.0,0.0);
  glVertex3f(1.0,1.0,0.0); glVertex3f(-1.0,1.0,0.0); glEnd();
#endif
#ifdef VERTICAL
  glBegin(GL_LINES); glVertex3f(-1.0,0.0,0.0); glVertex3f(1.0,0.0,0.0);glEnd();
  
  glRasterPos3f( -0.97,  0.96, 0.0 );
  printstring( GLUT_BITMAP_TIMES_ROMAN_10, "Camera 1 - Principal View");
  glRasterPos3f( -0.97, -0.04, 0.0 );
  printstring( GLUT_BITMAP_TIMES_ROMAN_10,
	"Camera 2 - Position and Viewing Volume of Camera 1");
#endif
#ifndef VERTICAL
  glBegin(GL_LINES); glVertex3f(0.0,-1.0,0.0); glVertex3f(0.0,1.0,0.0); glEnd();
  glRasterPos3f( -0.97, 0.94, 0.0 );
  printstring( GLUT_BITMAP_TIMES_ROMAN_10, "Camera 1 - Principal View");
  glRasterPos3f(  0.03, 0.94, 0.0 );
  printstring( GLUT_BITMAP_TIMES_ROMAN_10,
	"Camera 2 - Position and Viewing Volume of Camera 1");
#endif
  glPopAttrib();
  /* restore the current view and projection */
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}


void display( void )
{
  glClearColor( .9f, .9f, .9f, 1.0f );
  glClearColor( .7f, .7f, .7f, 1.0f );
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


  // draw view #1 (Principal view)
  draw_view1();
  draw_info();
  // draw view #2 (Second observer that looks at everything)
  draw_view2();

  glutSwapBuffers();
}


void reshape( int wid, int hei )
{
  w = wid; h = hei;
}


static void initLights() {
  /****************************************/
  /* Set up OpenGL lights     */
  /****************************************/

  GLfloat light0_ambient[] =  {1.0f, 1.0f, 1.0f, 1.0f};	// White
  GLfloat light0_diffuse[] =  {1.0f, 1.0f, 1.0f, 1.0f};	// White
  //GLfloat light0_ambient[] =  {0.0f, 0.0f, 1.2f, 1.0f};	// Blue
  //GLfloat light0_diffuse[] =  {0.0f, 0.0f, 1.2f, 1.0f};	// Blue
  GLfloat light0_position[] = {1.0f, 1.0f, 4.0f, 0.0f};

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
}

 
//#define SCEN2
#define SCEN3
// Put some special parameters for producing special output
static void docuSetup() {
  Reset(0);
  wireframe = (int)GL_TRUE; vvwireframe = (int)GL_TRUE;
  object_type = 0; // 0=Teapot 1=Cube 2=Sphere 3=Cone 4=Cylinder 5=Torus
  //top2 = 1.0; 	// Force volume to touch upper limit


  //camera 1
  eyex = 0.0; eyey = 0.0; eyez = 5.0;
  centerx = 0.0; centery = 0.0; centerz = 0.0;
  upx = 0.0; upy = 1.0; upz = 0.0;
  left = -1.5; right = 1.5; bottom = -1.5; top = 1.5;
  near_ = 3.0; far_ = 6.5;
  fovy = 20.0; aspect = 4.0/3.0;
  orthographic = 0; 
  frustum = 1;  

  //camera 2
  eyex2 = 20.0; eyey2 = 10.0; eyez2 = 10.0; 
  centerx2 = 0.0; centery2 = 0.0; centerz2 = 0.0; 
  upx2 = 0.0; upy2 = 1.0; upz2 = 0.0;
  left2 = -2.0; right2 = 2.0; bottom2 = -2.0; top2 = 2.0; 
  near_2 = 8.0; far_2 = 40.0;
  fovy2 = 20.0; aspect2 = 4.0/3.0;
  orthographic2 = 0;
  frustum2 = 1; 
#ifdef SCEN2
//
// Parameters of cam 2 for 2D:
// 0,20,0, 0,0,0, 0,0,-1, ortho, 10,40, -4,4, -6,2
//
  object_type = 3; // 0=Teapot 1=Cube 2=Sphere 3=Cone 4=Cylinder 5=Torus
  left = -0.4;
  //camera 2
  eyex2 = 0.0; eyey2 = 20.0; eyez2 = 0.0; 
  centerx2 = 0.0; centery2 = 0.0; centerz2 = 0.0; 
  upx2 = 0.0; upy2 = 0.0; upz2 = -1.0;
  near_2 = 10.0; far_2 = 40.0;
  left2 = -4.0; right2 = 4.0; bottom2 = -6.0; top2 = 2.0; 
  orthographic2 = 1; frustum2 = 0; 
#endif
}

/**************************************** main() ********************/

int main(int argc, char* argv[])
{
  /****************************************/
  /*   Initialize GLUT and create window  */
  /****************************************/

  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowSize( w_init, h_init );
  glutInitWindowPosition( 100, 100 );
  // Do not use, had problems with it // glutSetIconTitle( "VVV" ); 
  glutInit( &argc, argv );
#ifndef FEEDBACK
  main_window =
#endif
	glutCreateWindow( "VIEWING VOLUME VISUALIZATION --- Thomas W. Rauber" );
#ifndef FEEDBACK
  glutDisplayFunc( display );
  initLights();
#endif
#ifdef FEEDBACK
  docuSetup();
  initEpsdriver();
  glutDisplayFunc(displayFeedback);
  glutKeyboardFunc(keyboard);
#endif
  glutReshapeFunc( reshape );

  /****************************************/
  /*    Enable Z-buffer    */
  /****************************************/

  glEnable(GL_DEPTH_TEST);

#ifndef FEEDBACK
  create_glui_interface();
#endif
  glutMainLoop();
  return 0;
}
